/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <InstrumentManager/InstrumentManager.h>
#include <boost/lexical_cast.hpp>
#include <fstream>
#include <iostream>
#include <common/base/ServerProperties.h>
#include "FrequencyRemoteMatlabReq.h"
#include "FrequencyRequestMessages.pb.h"
#include "controllers/common/family/Families.h"

namespace in16 {

using namespace std;
using namespace common;
using namespace boost;
using namespace cameo;

const std::string FrequencyRemoteMatlabReq::TYPE = "in16_remote_matlab";
const std::string FrequencyRemoteMatlabReq::MATLAB_APPLICATION = "in16-matlab";
const std::string FrequencyRemoteMatlabReq::MATLAB_STOP_APPLICATION = "in16-matlab-stop";
const std::string FrequencyRemoteMatlabReq::RESPONDER_NAME = "in16_frequency_controller";

const std::string FrequencyRemoteMatlabReq::RUN_SPECTRUMS_TEST = "Run Spectrums Test";
const std::string FrequencyRemoteMatlabReq::RUN_SPECTRUMS = "Run Spectrums";
const std::string FrequencyRemoteMatlabReq::RUN_SPECTRUMS_TIME = "Run Spectrums Time";
const std::string FrequencyRemoteMatlabReq::RUN_SINGLE_FREQUENCY = "Run Single Frequency";

FrequencyRemoteMatlabReq::FrequencyRemoteMatlabReq(const std::string& name) :
	ExperimentController(name), controller::Start(this) {

	serverEndpoint.init(this, SAVE, "cameo_server");
	initialized.init(this, NOSAVE, "initialized");

	requestTypeValues.init(this, NOSAVE, "request_type_values");

	requestTypeValues.resize(4);
	requestTypeValues.set(0, RUN_SPECTRUMS_TEST);
	requestTypeValues.set(1, RUN_SPECTRUMS);
	requestTypeValues.set(2, RUN_SPECTRUMS_TIME);
	requestTypeValues.set(3, RUN_SINGLE_FREQUENCY);

	requestType.init(this, NOSAVE, "request_type", "Request");
	requestType.setEnumeratedLabels(requestTypeValues);
	requestType.setEnumeratedValues(requestTypeValues);

	requestType = requestTypeValues.get(0);

	fr.init(this, NOSAVE, "fr");
	frMin.init(this, NOSAVE, "frMin");
	frMax.init(this, NOSAVE, "frMax");
	nMax.init(this, NOSAVE, "nMax");
	time.init(this, NOSAVE, "time");
	comment.init(this, NOSAVE, "comment");

	fr = 0;
	frMin = 1E1;
	frMax = 1E6;
	nMax = 5;
	time = 0;
	comment = "a comment";
}

FrequencyRemoteMatlabReq::FrequencyRemoteMatlabReq(const FrequencyRemoteMatlabReq& controller) :
	ExperimentController(controller), controller::Start(this) {
}

FrequencyRemoteMatlabReq::~FrequencyRemoteMatlabReq() {

	// Stop the remote application.
	m_matlabApplication->stop();

	// Wait for the termination
	application::State state = m_matlabApplication->waitFor();

	cout << "Matlab application terminated with state " << application::toString(state) << endl;
}

void FrequencyRemoteMatlabReq::postConfiguration() {

	if (serverEndpoint() == "") {
		m_server.reset(new Server(application::This::getServer().getEndpoint()));

	} else {
		m_server.reset(new Server(serverEndpoint()));
	}

	initialized = initApplication();
}

bool FrequencyRemoteMatlabReq::initApplication() {

	// Do not initialize if it is already done
	if (initialized()) {
		return true;
	}

	try {

		// Start the Matlab server
		if (!m_server->isAvailable(10000)) {
			cout << "Matlab server is not available" << endl;
			return false;
		}

		cout << "Matlab server is connected to " << getName() << endl;

		m_matlabApplication = m_server->connect(MATLAB_APPLICATION);
		if (m_matlabApplication->exists()) {
			// The application exists from a previous server session
			m_matlabApplication->stop();
			application::State state = m_matlabApplication->waitFor();
			cout << "Terminated matlab application " << state << endl;
		}

		m_matlabApplication = m_server->start(MATLAB_APPLICATION);

		if (!m_matlabApplication->exists()) {
			cout << "No matlab application" << endl;
			return false;
		}

		// Create the requester.
		m_requester = application::Requester::create(*m_matlabApplication, RESPONDER_NAME);

		if (m_requester.get() == 0) {
			cout << "requester error" << endl;
			return false;
		}

		// Application initialized.
		initialized = true;

		return true;

	} catch (...) {
		// Currently an exception can occur during isAvailable.
		cout << "Unexpected exception during matlab connection" << endl;
	}

	return false;
}

void FrequencyRemoteMatlabReq::start() {

	// Serialize the request.
	proto::RunSpectrumsRequest request;

	request.set_fr(fr());
	request.set_frmin(frMin());
	request.set_frmax(frMax());
	request.set_nmax(nMax());
	request.set_time(time());
	request.set_comment(comment());

	if (requestType() == RUN_SPECTRUMS) {
		request.set_type(proto::RunSpectrumsRequest_Type_RUN);
	}
	else if (requestType() == RUN_SPECTRUMS_TIME) {
		request.set_type(proto::RunSpectrumsRequest_Type_RUN_TIME);
	}
	else if (requestType() == RUN_SINGLE_FREQUENCY) {
		request.set_type(proto::RunSpectrumsRequest_Type_RUN_SINGLE);
	}
	else if (requestType() == RUN_SPECTRUMS_TEST) {
		request.set_type(proto::RunSpectrumsRequest_Type_TEST);
	}

	// Send the message.
	m_requester->sendBinary(request.SerializeAsString());

	cout << getName() << " sent run request " << requestType() << endl;

	// Wait for the response synchronously.
	// Note that responses can be processed asynchronously if the server is able to.
	string data;
	m_requester->receiveBinary(data);

	proto::FrequencyResponse response;
	response.ParseFromString(data);

	cout << "received response " << response.responseid() << endl;

	sleep(1);
}

void FrequencyRemoteMatlabReq::stopRun() {

	auto_ptr<cameo::application::Instance> matlabStopApplication = m_server->start(MATLAB_STOP_APPLICATION);
	application::State state = matlabStopApplication->waitFor();
	cout << "Matlab stop application terminated with state " << state << endl;
}

}
