/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CAENCDDRIVER_H
#define CAENCDDRIVER_H

#include <Driver.h>
#include "drivers/global/Vme.h"

namespace caen_cd {

/*!
 * \class CAENCdDriver
 * \brief Main class for the CAENCdDriver device driver
 *
 * \par
 * Class for CAENCdDriver driver. Implements DeviceDriver methods.
 * \par
 * The CAENCdDriver device manages the global functionality of the CAEN V1740 VME Digitiser.
 */

class CAENCdDriver : public driver::Vme {

public:

	//! Driver type value
	static const std::string TYPE;

	/*!
	 * \brief Constructor
	 * \param[in] name the name of the device driver
	 */
	CAENCdDriver(const std::string& name);

	/*!
	 * \brief Destructor
	 */
	virtual ~CAENCdDriver();

	/*!
	 * \brief Method called for executing a command
	 *
	 * \param[in] command the command to apply on the controller
	 */
	virtual void execute(const std::string& aCommand);

	// Global properties
	Property<int32> nbChannels;
	Property<int32> nbBits;
    Property<int32> model;
    Property<int32>	channels;
    Property<int32> serialNumber;
    Property<std::string> modelName;
    Property<std::string> rocFirmware;
    Property<std::string> amcFirmware;

    // Board properties
    Property<int32> linkType;
    Property<int32> acqMode;
    Property<int32> dppAcqMode;
    Property<std::string> dppAcqModeStr;
    Property<int32> saveMode;
    Property<std::string> saveModeStr;
    Property<int32> recordLength;					// Nb samples for oscilloscope mode
    Property<int32> ioLevel;
    Property<int32> triggerMode;
    Property<int32> channelMask;

	// Channels properties in dppParams - PSD
    Property<int32>triggerHoldOff;					// trgho - same for all channels
    DynamicProperty<int32>triggerThreshold;			// thr (N x 4 ns)
    DynamicProperty<int32>selfTrigger;				// selft - range 0:1 (disable - enable)
    DynamicProperty<int32>chargeSensitivity;		// csens - range 0:3 (40fc/LSB, 160fc/LSB, 640fc/LSB, 2,5pc/LSB)
    DynamicProperty<int32>shortGate;				// sgate (N x 4 ns)
    DynamicProperty<int32>longGate;					// lgate (N x 4 ns)
    DynamicProperty<int32>preGate;					// pgate (N x 4 ns)
    DynamicProperty<int32>triggerValidationWindow;	// tvalw
    DynamicProperty<int32>baselineMean;				// nsbl - range 0:6 (no baseline, 4, 8, 16, 32, 64, 128)

	// Channels properties outside dppParam
	DynamicProperty<int32> channelActive;
	DynamicProperty<std::string> channelActiveStr;
	DynamicProperty<int32> dcOffset;
	DynamicProperty<int32> pulsePolarity;			// invert input - range 0:1
	DynamicProperty<std::string> pulsePolarityStr;
	DynamicProperty<int32> preTrigger;				// range 0:1023 (nSamples = pre_trig x 2)

	// Commands
	static const std::string WRITE_PARAMETERS_COMMAND;
	static const std::string SET_MODE_COMMAND;
	static const std::string RESET_COMMAND;
	static const std::string CLEAR_COMMAND;

private:

	void refreshNbChannelsProperty(int32 aValue) throw (CannotSetValue);
	void refreshDppAcqModeProperty(int32 aValue) throw (CannotSetValue);
	void refreshSaveModeProperty(int32 aValue) throw (CannotSetValue);
	void refreshChannelActiveProperty(int32 index, int32 aValue) throw (CannotSetValue);
	void refreshPulsePolarityProperty(int32 index, int32 aValue) throw (CannotSetValue);
};

}
#endif //CAENCDDRIVER_H
